/* bhole header - main stuff */
#ifndef _BHOLE_H_MAIN
#define _BHOLE_H_MAIN

#include<stdio.h>
#include<string.h>
#include<malloc.h>
#include<time.h>
#include<SDL.h> /* yeah baby, SDL */
#include<stdlib.h>
#include<math.h>
#include<limits.h>
#include "SDL_endian.h"

#ifdef WIN32
#include<windows.h>
#include<io.h>
#include<ctype.h>
#else
#include<ctype.h>
#endif

#ifndef QUZAR
#include<sys/stat.h>
#include<sys/time.h>
#endif

#include<unistd.h>
/*#else
#include<winsock.h>
#endif
*/

extern FILE *report_file;

/**************** HERE COME ASSUMPTIONS ******************/
/* assumptions are things which are not standard
 * and are not used by all 80386 DOS programs
 * this assumptions guarntee speed
 * and sometimes, you gotta have them in order to run the program
 * one possible example of this is ASSUME_CGA
 * all assumations except for assume_com_file are #defined for
 * maximum speed...
 * (they are hard-coded, not soft-coded)
 */
/* assume_video_page_flipping means:
 * BHole assumes the program ALWAYS uses page-flipping MODE-X graphics
 * and only synchronizes screen redraw with page-flipping
 *
 * game needing this: Master Of Magic
 */
#define ASSUME_VIDEO_PAGE_FLIPPING 
#undef  ASSUME_VIDEO_PAGE_FLIPPING

/* ASSUME_CGA means:
 * BHole acts like the video card is a CGA, thus allowing the old and obsolete
 * 160x100 16 colors resolution mode
 *
 * game needing this: round 42, moonbugs
 */
#define ASSUME_CGA

/* ASSUME_EGA means:
 * BHole will be availble for EGA mode 0xD
 *
 * games needing this: duke nukem
 */
#define ASSUME_EGA

/* assume_int1c_as_irq means:
 * interrupt 0x1C is called as an IRQ every time handle_events
 * is called (notice: this makes an application think time is slow/fast/wierd)
 * 
 * game needing this: round 42, moonbugs, livingstone
 * 
 * notice: in fact, the real IRQ is interrupt 0x08 which calls interrupt 0x1C
 * notice: by enabling this, int 08 as IRQ is automatically enabled as well
 */

#define ASSUME_INT1C_AS_IRQ

/* assume_int9_as_irq means:
 * interrupt 0x9 is called as an IRQ every time a key is pressed/released
 * 
 * game needing this: round 42, moonbugs
 * 
 * NOTICE: TODO: emulated program cannot disable the int 16 BIOS
 *               service by chaining in here
 */
#define ASSUME_INT9_AS_IRQ

/* assume_direct_mode3 means:
 * this program uses graphical mode 3 directly and outputs stuff
 * to B800 - games needing this: round 42 opening screen
 */
#define ASSUME_DIRECT_MODE3
#undef  ASSUME_DIRECT_MODE3






/* HYPER_SPEED_MODE - in this mode
 * everything will work in a more efficient way
 * but some features will be disabled:
 * effects of HYPER_SPEED_MODE:
 * - debug_vcpu will be ignored
 * - get_csip will IGNORE memory redirection
 * - in little endian systems, get_csip_word
 *   will work faster
 * - IP will be stored in the size of the default machine
 *   (for pentium, 32-bit) instead of 16-bit
 */
#define HYPER_SPEED_MODE


/* system sizes check */
/* TODO: modify the code to handle these kind of shits */
#if USHRT_MAX != 65535
#error USHRT_MAX does not fit
#endif
#if UINT_MAX != 4294967295U
#error UINT_MAX does not fit
#endif
#if UCHAR_MAX != 255
#error UCHAR_MAX does not fit
#endif
#if ULONG_MAX != 4294967295UL
#error ULONG_MAX does not fit
#endif
extern int assume_com_file ;
extern int assume_img_file ;

#define REGSCOUNT 12
#define GREGSCOUNT 8
#define SREGSCOUNT 4
#define REGSSIZE  24
#define MEMSIZE 1048576


extern unsigned char __emustuff__[REGSSIZE + MEMSIZE];
extern char *execname;
extern FILE *execfile;

#define regs ((unsigned short*)__emustuff__)

#define ax regs[0]
#define cx regs[1]
#define dx regs[2]
#define bx regs[3]
#define sp regs[4]
#define bp regs[5]
#define si regs[6]
#define di regs[7]
#define es regs[8]
#define cs regs[9]
#define ss regs[10]
#define ds regs[11]

#define al *((unsigned char*)&ax)
#define cl *((unsigned char*)&cx)
#define dl *((unsigned char*)&dx)
#define bl *((unsigned char*)&bx)
#define ah *((unsigned char*)&ax+1)
#define ch *((unsigned char*)&cx+1)
#define dh *((unsigned char*)&dx+1)
#define bh *((unsigned char*)&bx+1)


#define IS_REG_PTR(p) ( ((void*)regs <= (p)) && ((p) <= (void*)(regs+11)) )




#ifndef HYPER_SPEED_MODE
extern unsigned short ip;
#else
extern unsigned int   ip; /* unless program overflows IP all should be okay i hope */
#endif


extern unsigned short flags;

#define GET_CF() (flags&0x01)
#define SET_CF() flags|=0x01
#define CLEAR_CF() flags&=(~0x01)
#define UNSET_CF() CLEAR_CF()

#define GET_PF() (flags&0x04)
#define SET_PF() flags|=0x04
#define CLEAR_PF() flags&=(~0x04)
#define UNSET_PF() CLEAR_PF()

#define GET_AF() (flags&0x10)
#define SET_AF() flags|=0x10
#define CLEAR_AF() flags&=(~0x10)
#define UNSET_AF() CLEAR_AF()

#define GET_ZF() (flags&0x40)
#define SET_ZF() flags|=0x40
#define CLEAR_ZF() flags&=(~0x40)
#define UNSET_ZF() CLEAR_ZF()

#define ZF_ON  0x40


#define GET_SF() (flags&0x80)
#define SET_SF() flags|=0x80
#define CLEAR_SF() flags&=(~0x80)
#define UNSET_SF() CLEAR_SF()

#define GET_OF() (flags&0x8000)
#define SET_OF() flags|=0x8000
#define CLEAR_OF() flags&=(~0x8000)
#define UNSET_OF() CLEAR_OF()


#define GET_DF() (flags&0x0400)
#define SET_DF() flags|=0x0400
#define CLEAR_DF() flags&=(~0x0400)
#define UNSET_DF() CLEAR_DF()


/* 0x000 - 0x3FF interrupt vector */

/* should ENVPOS be 0x9DE0??? */

#define ENVPOS (0x09CF0)
#define PSPPOS (0x119E0)
#define PSPSEG (0x119E)
/* #define DTA_OFFSET (0x80) - OBSOLETE  PSPSEG:DTA_OFFSET = address of DTA */
/* dta is no longer constant */
/* dta is assumed to be in dta_segment:dta_offset and is set
 * in the routine init_env to PSP:0x80 
 */
unsigned short dta_offset;
unsigned short dta_segment; 

/*
 * THESE MACROS WERE CANCELLED CAUSE THEY ALLOW
 * DIRECT MEMORY ADDRESS
 *
 * #define ENV (memory + ENVPOS)
 * #define PSP (memory + 0x119E0)
 *
 */

#define RESERVED 0x11AE0 /* size of reserved stuff before prog can be */

#define __memory__ (__emustuff__ + REGSSIZE)

/* is this definition safe??? */
#define pseudoclock (*(long*)(__memory__ + 0x46C))

unsigned int line;

#define emumem2mem(mem,seg,index,len) emudmem2mem((mem),((seg)<<4)+(index),(len));
#define mem2emumem(seg,index,mem,len) mem2emudmem(((seg)<<4)+(index),(mem),(len));
void  emudmem2mem(unsigned char *mem,unsigned int offset,int len);
void  mem2emudmem(unsigned int offset,unsigned char *mem,int len);




/* this declares the important routine used
 * to redirect memory correctly
 */
unsigned char *_redirected_addr(unsigned int   ofs);
unsigned char  get_vidmem(int doffset);
void set_vidmem(unsigned int doffset,unsigned char value);


#define IS_MEMORY_PTR(ptr) ( (__memory__ <= ((unsigned char*)(ptr))) && (((unsigned char *)(ptr)) < __memory__ + 1048576) )
#define IS_CLOCK_PTR(ptr) (__memory__ + 0x470 > (ptr) && (ptr) >= __memory__ + 0x46C)
#define IS_BELOW_HIGHMEM_PTR(ptr) ( ((unsigned char*)(ptr)) < __memory__ + 0xA0000 )
#define IS_BELOW_HIGHMEM_OFS(ofs) (                  (ofs)   <              0xA0000 )
/* should vidmem be considered as 64000 or as 65536??? */
#define IS_VIDMEM_OFS(ofs) ( (           + 0xA0000 <= ofs                    ) && (ofs                      <            + 0xA0000 + 65536 ) )
#define IS_BELOW_VIDMEM_LIMIT_OFS(ofs) ((ofs) < 0xA0000 + 65536)

#define IS_BELOW_FULLVIDMEM_LIMIT_OFS(ofs) ((ofs) < 0xC0000)

#define redirected_addr(ptr) (_redirected_addr(ptr))

#define SET_MEM_BYTE(seg,offset,value) SET_MEM_BYTE2((offset) + ((seg) << 4),value)
#define GET_MEM_BYTE(seg,offset      ) GET_MEM_BYTE2((offset) + ((seg) << 4))


/************* OLD UNUSED VERSION
#define GET_MEM_BYTE2(doffset) \
(IS_VIDMEM_OFS(doffset) ? get_vidmem(doffset) : ( *((unsigned char*)(redirected_addr(__memory__ + (doffset)  ) ) ) ) ) \
**************/

#if defined(ASSUME_VIDEO_PAGE_FLIPPING) || defined(ASSUME_EGA)

#ifdef CHECK_INLIM_MEM
#error unimp
#endif

#define GET_MEM_BYTE2(doffset) \
(IS_BELOW_HIGHMEM_OFS(doffset) ? *(__memory__ + (doffset)) : IS_BELOW_VIDMEM_LIMIT_OFS(doffset) ? get_vidmem(doffset) : \
 *(redirected_addr(doffset)))

#else

#ifndef CHECK_INLIM_MEM

#define GET_MEM_BYTE2(doffset) \
(IS_BELOW_FULLVIDMEM_LIMIT_OFS(doffset) ? *(__memory__ + (doffset)) : *(redirected_addr(doffset)))

#else

#define GET_MEM_BYTE2(doffset) \
(IS_BELOW_FULLVIDMEM_LIMIT_OFS(doffset) ? *(__memory__ +  check(doffset)) : *(redirected_addr(doffset)))

int check(int dofs)
{
	if (dofs < CHECKLIM1 && dofs > CHECKLIM2)
	{
		printf("*_addCKLIM1 && dofs > CHECKL3
	if \;dCKLIM1 && dofs > CHECKL3
	if \;d,routinr,E}r,EfsER_nIM1 &nr,}r, \
(IS_BELOW
(IS_BELO******************/
doffset) \
(IS_VIDMEM_OFS(dSffset) ? get_vidmem(    ) Gf  ? {						\2)
	{
t) : ( *((unsigned chaGf  ? f s(doffset)))

#e    ) G;  ? f(IS_					\2)
	ory__ + (doffset)  ) ) ) ) ) \
**************/

#if define  <                ER_S  ) GG;  ? }						\2)*gs which are not st kind of shSffset) ? get,ed a1 && dos 65536?
#e__ (__e,??? */
1 &&dofo me
#dtr && dosNG) || defined(ASSUME_EGA)

#ifdef CHECK_INLIM_MEM
#error unimp
#endif

#define GET_MEM_BYTE2(doffset)leED Ced \
(IS_BELOW_HIGHMESffset) ? get_vidmem(    ) Gf 	f  ? {									\2)
	{
t) + (doffset)) : IS_BELOW_VI)				\2)
{								\2)
	*_ + (doffset)  ) *********/

#if defin  ER_S  ) G;  ? f}							  ? f(IS_ 
	{
t) + (dof  check(doffset)) : *(red)			\2)
{								\2)
	s(doffset)))

#e    ) G; 		  ? f}							  ? f(IS_			 				\2)
{								\2)
	*_ + (doffset)  ) t dofs)
{
	if (dofs < C ER_S  ) G;  ? f}							  ? }							 	\efine GET_MEM_BYTE2(doffset) \
(IS_BELOW_FULLVISffset) ? get_vidmem(    ) Gf 	f  ? {									\2)
	{
t) + (doffset)) : IS_BELOW_VI)				\2)
{								\2)
	*_ + (doffset)  ) *********/

#if defin  ER_S  ) G;  ? f}							  ? f(IS_			 				\2)
{								\2)
	*_ + (doffset)  ) t dofs)
{
	if (dofs < C ER_S  ) G;  ? f}							  ? }							 	\efine GET_MEMLOW_FULLVISffset) ? get_vidmem(    ) Gf 	f  ? {									\2)
	{
t) + (doffset)) : IS_BELOW_VI)				\2)
{								\2)
	
	{
BELOW_V    tf("*_addCKLIM1 OW_V >  tf("*_ad2)				\2)
f \;dCKLIMS && dofs > CHECKL3
	if  ->if \;d,   ) ,ffset,uns	\2)
	*_ + (doffset)  ) *********/

#if defin  ER_S  ) G;  ? f}							  ? f(IS_			 				\2)
{								\2)
	*_ + (doffset)  ) t dofs)
{
	if (dofs < C ER_S  ) G;  ? f}							  ? }							 	\efin \
(IS_BELOW
(IS_BELOind HE FOfine
/* DIRECTAEMODISABRESSS
C
 * #dDOable
ENV 
 * andefiney + ENVPEM(memor_VIDMe of rBELOW_FULLVIDMEM_LIMWEND+ ((seg) <<t) ? * * ax regs[0]
#defiERVED 0x11** OLD UNUSED VERSION
 tern unsigW_FULLVIDMEM_LIMDWEND+ ((seg) <<t) ? * * ax regs[0]inefiERVED 0x11** OLD UNUSED VERSION
 tern unsigW_FULLVISMEM_LIMWEND+ ((seg) <<,   ) t) ? * * ax regs[0]
#defiERVED 0x11** OLD UNUSED VERSION
 terER_S  ) Grn unsdefine END OFORY ADDAr_VI ZVIEigned iBELOW_FULLVIDMEM_LIMWEND+ ((seg) <<t) ? *                                 ?  e ax regs[0]
#de)E2((offset) + ((seg) <<ine  <    UNUS ?  e                 E2((offset) + ((s** OLD U+1RSION8)    ? )nsigW_FULLVISMEM_LIMWEND+ ((seg) <<,   ) t)                 ?  {                                                      ?  SE MACROS WERE     ER_S  ) G;                            ?  E2((offset) + ((seg) <<  ,        &(_r beeeeeeeee);    ?  E2((offset) + ((seg) <<+1mum      &(_r brec >>N8));    ?  }                                                     \BELOW_FULLVIDMEM_LIMWENDFS(doffset) ? *                                 ?  e ax regs[0]
#de)E2((offset) FS(doffset)))))))))NLIS ?  e                 E2((offset) ***d* OLD U+1RSION8)    ? )nsigW_FULLVISMEM_LIMWENDt_vidmem(    ) G                         ?  {                                                           ?  SE MACROS WERE     ER_S  ) G;                               ?  E2((offset) t_
#if define  <  ,         &(_r beeeeeeeee);  ?  E2((offset) t_**d* OLD U+1RS  ,((        &(_r brec >>N8));  ?  }